/**********************************************************************/
/** Copyright (C) 2021 - Université Catholique de Louvain		**/
/** affinistan version 2021.1						**/
/** Author: Mohamed HADDAD (immc-MEMA)				**/
/** Contact: mohamed.haddad@uclouvain.be				**/
/** Use of this source code is governed by an UCL-style license	**/ 
/** All Rights Reserved						**/
/** This project has received funding from the European 		**/
/** Union’s Horizon 2020 research and innovation program		**/
/** under grant agreement No 862015					**/
/***********************************************************************/

#ifndef FONCTIONS_AFFINE_H
#define FONCTIONS_AFFINE_H 1

#include "matprop.h"
#include "state.h"

void readinput (MATPROP *matr, MATPROP *incl);
void initialstate(MATPROP *matrprop, MATPROP *inclprop, STATEMACRO *macrot0, STATE *matrt0, STATE *inclt0, double **lim1H, double ***lim);
void computestrslinel(double young, double nu, double *strn, double *strs);
void computestrnlinel(double young, double nu, double *strs, double *strn);
double getstrnrate(double tn);
double getdt();
void getdstrn(double strnrate, double dt, double nu, double *dstrn);
double resuniax (double *strs);
void corruniax (double *strs, double **C, double *corrstrn);
void compute_esh (double **C0, double ar, double **Esh);
void compute_hill(double **C0, MATPROP *inclprop, double **P); 
void compute_strn_conc(double **C0, double **C1, MATPROP *inclprop, double **H);
void compute_thstrn_conc(double **C0, double *eigenstrn0, double **C1, double *eigenstrn1, double v1, double **H,double **F,double **G);
void homog_tens_thel(double **C0, double *alpha0, double **C1, double *alpha1, double v1, double **H,\
                       double **Cmacro, double *eigenstrnmacro);
void inverse_Laplace (int m, int n, double ***Fs, double *s, int numbercolloc, double **lim1Fs, double **lim2Fs, int choice, double ***btau);
void eval_temp(int m, int n, double ***btau, double *s, int nbrtimes, double *t, int choice, double **lim1Fs, double **lim2Fs, double ***Ft);
void eval_Laplace(int m, int n, double ***btau, double *s, int nbreval, double *seval, int choice, double **lim1Fs, double **lim2Fs, double ***Fs);
void compute_R (double p,int chhard,double hardmod,double hardexp, double yldstrs, double *Rp, double *dRdp);
void computevisc (double strseq, double p, MATPROP *mat, double *gv, double *dgvdstrseq, double *dgvdp);
void convolution_series(double **A, double **B, double ***btau, int numbercolloc, double *s, int chinterp, double *g0, double *gtn, double *g, double tn, double dt, double *intgtn, double **intgexpotn, double *intg, double **intgexpo,double *result);
void convseries2 (double tn, double dt, int numbercolloc, double *s, int choiceinterpC, int choiceinterpeigenstrn, \
                  double **A, double **B, double ***bk, double *C, double *D, double ***dk, \
                  STATEMACRO *macrot0, STATEMACRO *macrotn, STATEMACRO *macrotau);
void convseriesdirect(double **lim1C, double **lim2C, double ***btauC, int choiceinterpC, double *lim1eigenstrnrate, double *lim2eigenstrnrate, double ***btaueigenstrn, int choiceinterpeigenstrn, double *dstrn, double timeincr, double tau, double *s, double *strs);
void limtimeindpt (MATPROP *matr, MATPROP *incl, double **lim1H, double ***lim);
void lim1aeps(MATPROP *matr,MATPROP *incl, double **lim1H, double *Cmstrs0, double *Cmstrs1, double ***lim);
void lim1eigenstrnratemacro(MATPROP *matr,MATPROP *incl, double *Cmstrs0, double *Cmstrs1, double ***lim);
void mallocstate(STATE *st);
void freestate(STATE st);
void copystate(STATE *A, STATE *B);
void fprintstate(FILE *pfile, STATE *A);
void printstate(STATE *A);
void mallocstatemacro (STATEMACRO *st);
void freestatemacro (STATEMACRO st);
void copystatemacro(STATEMACRO *A, STATEMACRO *B);
void fprintstatemacro(FILE *pfile, STATEMACRO *A);
void printstatemacro(STATEMACRO *A);
void fprintmatprop(FILE *pfile, MATPROP *A);
void printmatprop(MATPROP *A);
void fprintres(FILE *pfile, double time, STATE *st);
void fprintresmacro(FILE *pfile, double time, STATEMACRO *stmacro);
void compute_Ivol_Idev (double **I_vol, double **I_dev);
void G_tilde(MATPROP *matr, double dt, double *sum);
void K_tilde (MATPROP *matr, double dt, double *sum);
void ve_dstrn_vol_dev(double *ve_dstrn, double *vol_part, double *dev_part);
void E_inf_tilde(MATPROP *matr, double dt, double **E_inf, double **E_tilde);
void compute_si(MATPROP *matr, STATE *tn, double dt, double *dstrain_ve_dev, double **Si);
void compute_sigma_Hj(MATPROP *matr, STATE *tn, double dt, double *dstrain_ve_vol, double **sig_Hj);
void compute_sum_exp_si(MATPROP *matr, double **Si, double dtn_plus1, double *sum_si);
void compute_sum_exp_Sigma_Hj(MATPROP *matr, double **Sigma_Hj, double dtn_plus1, double *sum_Sigma_Hj);
void J_t(MATPROP *matr, double t, double **J);
void J_s(MATPROP *matr, double s, double **J);
void printmatrix(double **M, int nbr_rows);
void lim1Js(MATPROP *material, double **lim1Js);
void lim1Jprimet(MATPROP *material, double **lim1Jpt);
void lim1Gs(MATPROP *material, double **lim1Gs);
double compute_Gt(MATPROP *matr, double t);


#endif
