#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import scipy as spy
import scipy.stats as spystats

import os,shutil, fnmatch, pickle
import sys

from sys import exit
import pdb
import pandas as pd
import glob

import matplotlib.pyplot as plt
from scipy.spatial import distance
from scipy.interpolate import interp1d,interp2d
from functools import reduce

#==============================================================================
# determine the cross-covariance and the cross-correlation matrices
#==============================================================================
def RFCorrMat(nVar,nDim,ntau,dtau,funccorr,C0_XY,lc):    
    nxyz = ntau[0]*ntau[1]*ntau[2]
    # correlation length matrix for two parameters
    lList = [[0]*nDim for i in range(nxyz)]             # [nxyz][3,]
    
    count = -1
    C_XY = np.zeros((nVar**2,ntau[0],ntau[1],ntau[2]))
    R_XY = np.zeros((nVar**2,ntau[0],ntau[1],ntau[2]))
    for i in range(ntau[0]):
        for j in range(ntau[1]):
            for k in range(ntau[2]):
                count = count + 1
                tauList = [i*dtau[0], j*dtau[1], k*dtau[2]]                 # [taux, tauy, tau z]
                
                lList[count][0] = np.ones((nVar,nVar))*lc[0]        # lij_x = (nVar,nVar)
                lList[count][1] = np.ones((nVar,nVar))*lc[1]        # lij_y = (nVar,nVar)
                lList[count][2] = np.ones((nVar,nVar))*lc[2]        # lij_z = (nVar,nVar)
                
                lListx = lList[count][0]                        # (nVar, nVar)
                lListy = lList[count][1]
                lListz = lList[count][2]
                
                rx = np.zeros((nVar,nVar))                      # (nVar, nVar)
                ry = np.zeros((nVar,nVar))
                rz = np.zeros((nVar,nVar))
                
                rh = np.zeros((nVar,nVar))
                Ch = np.zeros((nVar,nVar))
                
                for a in range(nVar):
                    for b in range(nVar):
                        rx[a,b] = funccorr(tauList[0], lListx[a][b])      # r_ab,x = scalar
                        ry[a,b] = funccorr(tauList[1], lListy[a][b])
                        rz[a,b] = funccorr(tauList[2], lListz[a][b])
                        
                        # element-wise multiplication
                        rh[a,b] = rx[a,b]*ry[a,b]*rz[a,b]       # correlation component
                        Ch[a,b] = C0_XY[a,b]*rh[a,b]     # [nxyz](nVar, nVar)
                
                # element-wise multiplication            
                #rh = rx*ry*rz                   # (nVar, nVar)
                #Ch = C0_XY*rh                   # (nVar, nVar)
                
                ChVec = np.reshape(Ch,(Ch.shape[0]*Ch.shape[1],1))  # (nVar^2,)
                RhVec = np.reshape(rh,(rh.shape[0]*rh.shape[1],1))  # (nVar^2,)
                
                C_XY[:,i,j,k] = ChVec[:,0]                  # cross-covariance matrix
                R_XY[:,i,j,k] = RhVec[:,0]                  # cross-correlation matrix
    
    return C_XY

#==============================================================================
def RFCorrMatExpand(Ori_shape_D,R_rs,R_shape,R_tmp):
    # 1D model
    if (len(Ori_shape_D)==1):
        R_tmp[0:R_rs.shape[0], 0:R_rs.shape[1]] = R_rs      # (441, 10)
        mid_P1 = int(R_tmp.shape[1]/2)      # mid point = 9
        for N_rs in range(R_shape[0]):      # nVar*nVar
            for d1 in range(1,mid_P1+1):      # no. original grid points not count the 1st point
                R_tmp[N_rs, mid_P1+d1] = R_tmp[N_rs, mid_P1-d1]     # add extra discrete points --> (441, 18)
    
    # 2D model
    if (len(Ori_shape_D)==2):
        R_tmp[0:R_rs.shape[0], 0:R_rs.shape[1], 0:R_rs.shape[2]] = R_rs
        mid_P1 = int(R_tmp.shape[1]/2)
        mid_P2 = int(R_tmp.shape[2]/2)
        for N_rs in range(R_shape[0]):          # no. variables
    
            for d2 in range(mid_P2+1):          # no. y-coord. grid points
                for d1 in range(1,mid_P1+1):      # no. x-coord. grid points
                    R_tmp[N_rs, mid_P1+d1,d2] = R_tmp[N_rs, mid_P1-d1,d2]  
            for d1 in range(R_shape[1]):
                for d2 in range(1,mid_P2+1):
                    R_tmp[N_rs,d1, mid_P2+d2] = R_tmp[N_rs,d1, mid_P2-d2]
    
    # 3D model
    if (len(Ori_shape_D)==3):
        # assign the original R_rs to R_tmp --> 4th-order tensor
        R_tmp[0:R_rs.shape[0], 0:R_rs.shape[1], 0:R_rs.shape[2], 0:R_rs.shape[3]] = R_rs
        mid_P1 = int(R_tmp.shape[1]/2)          # x-dir
        mid_P2 = int(R_tmp.shape[2]/2)          # y-dir
        mid_P3 = int(R_tmp.shape[3]/2)          # z-dir
        #print(R_tmp.shape[1],mid_P1,mid_P2,mid_P3)
        for N_rs in range(R_shape[0]):          # no. variables
            for d3 in range(mid_P3+1):          # no. z-coord. grid points
                for d2 in range(mid_P2+1):      # no. y-coord. grid points
                    for d1 in range(1,mid_P1+1):  # no. x-coord. grid points
                        R_tmp[N_rs, mid_P1+d1, d2, d3] = R_tmp[N_rs, mid_P1-d1,d2,d3]
                        
                for d1 in range(R_shape[1]):
                    for d2 in range(1,mid_P2+1):
                        R_tmp[N_rs, d1, mid_P2+d2, d3] = R_tmp[N_rs,d1, mid_P2-d2,d3]
    
            for d1 in range(R_shape[1]):
                for d2 in range(R_shape[2]):
                    for d3 in range(1,mid_P3+1):
                        R_tmp[N_rs, d1, d2, mid_P3+d3] = R_tmp[N_rs,d1, d2, mid_P3-d3]

    R_rs=R_tmp      # (4,18,18,18)
    
    return R_rs

#==============================================================================
#H_pdf = Spectral_Analysis(R_rs, Grille_Dim)     # (18,18,18,4)
#==============================================================================
def RFSpecAnalysis(R_rs,Grille_Dim):
    frac_DtoC=1.0
    for i in range(len(Grille_Dim)):    # nDim
        if Grille_Dim[i] != 0.0 :
            frac_DtoC=frac_DtoC*Grille_Dim[i]   # \Delta \tau = 1/2B = 0.1
    
    Numb_Vsq = len(R_rs)        # no. variables squared = 4
    S_pdf = []
    for j in range(Numb_Vsq):   # nVar^2
        S_tmp = np.fft.fftn(R_rs[j]) * frac_DtoC    # power spectral density (18,); R_rs[j] = (18,) --> (18,18,18) for 3D
        S_pdf.append(S_tmp)                         # list [4](18,) --> [4](18,18,18) for 3D
    
    S_tmp = np.reshape(S_pdf,-1)                    # (72,) = 4*18 --> (23328,) = 4*18*18*18 for 3D
    ntauNew = [S_pdf[0].shape[0], S_pdf[0].shape[1], S_pdf[0].shape[2]]
    #ntauNew = [list_shape_D[0], list_shape_D[1], list_shape_D[2]]
    
    S_pdf = np.reshape(S_tmp,(ntauNew[0], ntauNew[1], ntauNew[2], Numb_Vsq), 'F')    # (18, 4) --> (18,18,18,4) for 3D
    #S_pdf = np.reshape(S_tmp,(2*ntau[0]-2, 2*ntau[1]-2, 2*ntau[2]-2, Numb_Vsq), 'F')    # (18, 4) --> (18,18,18,4) for 3D
    # S_pdf = (nx,ny,nz,nVar^2)
    
    # 3D model
    for i in range(R_rs.shape[1]):         # range(len(S_pdf.shape[0]))   <-- for i in range(len(S_pdf)):             # no. coord. grid points
        for j in range(R_rs.shape[2]):     # range(len(S_pdf.shape[1]))
            for k in range(R_rs.shape[3]): # range(len(S_pdf.shape[2]))
                # extract entities of S[m] (entities of the cross-covariance matrix) at the ith point in the freq. domain
                S_tmp0 = np.reshape(S_pdf[i,j,k],(int(len(S_pdf[0,0,0])**0.5),int(len(S_pdf[0,0,0])**0.5)))    # (nVar,nVar); S_pdf[i,j,k] = (nVar^2,) for 3D
                #S_tmp0 = np.reshape(S_pdf[i,j,k],(int(S_pdf.shape[3]**0.5),int(S_pdf.shape[3]**0.5)))    # (nVar,nVar); S_pdf[i,j,k] = (nVar^2,) for 3D
                #S_tmp0 = np.reshape(S_pdf[i],(int(len(S_pdf[0])**0.5),int(len(S_pdf[0])**0.5)))    # (21, 21) --> S_pdf[i,j,k] for 3D
                S_tmp = (S_tmp0 + S_tmp0.T)/2.0     # (nVar,nVar)
                #-------------------------------------
                # decomposition hermitian matrix S_tmp
                #-------------------------------------
                eigen_val,H_S = np.linalg.eig(S_tmp)    # eigen_val = (nVar,); H_S = (nVar,nVar)
                for r in range(len(eigen_val)):         # nVar
                    for s in range(len(eigen_val)):     # nVar
                        H_S[r][s] = H_S[r][s]*np.sqrt(eigen_val[s])
                # end decomposition hermitian
                
                S_pdf[i,j,k] = np.reshape(H_S,-1)   # S_pdf[i] = (441,) --> S_pdf[i,j,k] = (441,)
    
    # S_pdf (18,441) --> (18,18,18,441) for 3D
    return S_pdf
#==============================================================================
def RFCoeffs(Ori_shape_D,H_pdf,x,Grille_Dim,Gaussian,Nsim):
    H_pdf_save=np.array(H_pdf)                     # (18,441) --> (18,18,18,441) for 3D
    output=[]
    
    # for 3D model, liste_H_S = [18][18][18](nVar,nVar)
    # Note that H_pdf_save[i,j,k] = 4
    liste_H_S=[[[np.reshape(H_pdf_save[i,j,k], (int(H_pdf_save.shape[3]**0.5), int(H_pdf_save.shape[3]**0.5))) \
                 for k in range(H_pdf_save.shape[2])] for j in range(H_pdf_save.shape[1])] for i in range(H_pdf_save.shape[0])]  # list [18][18][18](nVar,nVar)
    
    Numb_F = tuple(H_pdf.shape[0:3])        # no. grid points in freq. domain = (18,18,18)
    Numb_V = len(liste_H_S[0][0][0])        # nVar
    Numb_D = len(Ori_shape_D)               # nDim = 3
    new_shape = Numb_F + (Numb_V,Numb_V)    # (18,18,18,nVar,nVar) for 3D
    
    Delta_F = []            # increment of frequency = \Delta \kappa_i
    index_F = []            # keep the N2*N3*...Nn, N3*...Nn,...,Nn
    total = np.asarray(Numb_F)              # no. freq. grid points = (18,18,18)
    frac = 2.0
    
    for m in range(Numb_D):                 # nDim = 3
        Delta_F.append(1/(Ori_shape_D[m]*Grille_Dim[m]))    # (3,); \Delta \kappa_i = 1/(2N_i - 1)/(\Delta \tau_i)
        total[m] = total[m]/Ori_shape_D[m]  # frac. between the no. grid points in spacial/time and freq. domains
        index_F.append(total[m])            # list [1](3,)
        frac = frac*Delta_F[m]              # 2 \Delta scalar
    
    alpha=np.sqrt(frac)                     # \sqrt{2 \Delta}
    
    liste_exp=[]
    test=[]
    for i in range(Numb_F[0]):              # no. freq. grid points in F1 dir.
        liste_exp.append([])
        test.append([])
        for j in range(Numb_F[1]):          # no. freq. grid points in F2 dir.
            liste_exp[i].append([])
            test[i].append([])
            for k in range(Numb_F[2]):      # no. freq. grid points in F3 dir.
                Coord_F=np.zeros(Numb_D)    # mi = (3,)
                i_mod = np.array([i, j, k]) # (3,)
                for m in range(Numb_D):     # nDim
                    Coord_F[m],i_mod[m] = divmod(i_mod[m],index_F[m])
                    if (Coord_F[m] > (Ori_shape_D[m])/2 ):    # make positive and negative frequency range
                        Coord_F[m] = Coord_F[m] - Ori_shape_D[m]# then the random value at the points between
                    test[i][j].append(Coord_F[m])               # [18](1,) # the sample points can be generated.
                # lopp on special position x
                exp = np.zeros(len(x), 'complex')   # (3,)
                for npoint in range(len(x)):        # no. coord. grid points in physical domain
                    # exp(2*pi*i*(\bm{x} \cdot \bm{\kappa}^(mxmymz)_m))
                    exp[npoint] =np.exp(2.0*np.pi*1j*np.sum(Coord_F*Delta_F*x[npoint]))  # (3,)
                
                liste_exp[i][j].append(exp)         # list [18](3,) --> [18][18][18](3,) for 3D
    
    #exit()
    for l in range(Nsim):
        print("Simulation", l)
        for i in range(H_pdf.shape[0]):                             # no. grid points in F1 dir.
            for j in range(H_pdf.shape[1]):                         # no. grid points in F2 dir.
                for k in range(H_pdf.shape[2]):                     # no. grid points in F3 dir.
                    #H_S = np.reshape(H_pdf[i],(len(H_pdf[0])**0.5,len(H_pdf[0])**0.5))
                    H_S = np.array(liste_H_S[i][j][k])              # (nVar,nVar); liste_H_S = [18][18][18](2,2)
                    
                    #add randomness in H_s----------
                    theta = np.random.uniform(0, 1, len(H_S))       # \theta(s,mx,my,mz)
                    rand = np.exp(2.0*np.pi*1j*theta)               # exp(2*pi*i*\theta(s,mx,my,mz))
                    if(Gaussian):
                        phi = np.random.uniform(0, 1, len(H_S))     # \phi(s,mx,my,mz)
                        rand = rand*np.sqrt(-np.log(phi))           # \eta(s,mx,my,mz)*exp(2*pi*i*\theta(s,mx,my,mz))
                    
                    for r in range(len(H_S)):                       # len(H_S) = 21
                        for s in range(len(H_S)):
                            H_S[r][s] = H_S[r][s]*rand[s]           # H^(rs)[\bm{m}^(mxmymz)]*\eta(s,mx,my,mz)*exp(2*pi*i*\theta(s,mx,my,mz)); H_S = (21, 21)
                    
                    H_pdf[i,j,k] = np.reshape(H_S,-1)               # H_pdf[i][j][k] = (nVar^2,); H_pdf = (18,18,18,4) for 3D
    
        # number of loop on frequency and number of variable in vector
        H_pdf = np.reshape(H_pdf,-1)           # (7938,)
        H_pdf = np.reshape(H_pdf,new_shape)    # (18, 21, 21) --> (18,18,18,nVar,nVar) for 3D
    
        # loop on frequency and variable 
        liste_sol=[]
        for i in range(Numb_F[0]):
            liste_sol.append([])
            for j in range(Numb_F[1]):
                liste_sol[i].append([])
                for k in range(Numb_F[2]):
                    liste_sol[i][j].append([])
                    for r in range(Numb_V):
                        tmp=0.
                        for s in range(Numb_V):
                            tmp += H_pdf[i,j,k,r,s]     # H_pdf = (18,18,18,nVar,nVar)
                        liste_sol[i][j][k].append(tmp)  # list [18][21](1,) --> [18][18][18][nVar](1,) for 3D
        
        rand_field = np.zeros((len(x),Numb_V), 'complex')  # (nCoord,nVar)
        for i in range(Numb_F[0]):
            for j in range(Numb_F[1]):
                for k in range(Numb_F[2]):
                    for npoint in range(len(x)):        # nCoord
                        # loop on variable
                        for r in range(Numb_V):         # nVar
                            rand_field[npoint][r] = rand_field[npoint][r] + liste_sol[i][j][k][r]*liste_exp[i][j][k][npoint]
        rand_field = alpha*rand_field   # (3, 21)
        
        H_pdf=np.array(H_pdf_save)     # (18, 441)
        
        output.append(rand_field.real)  # list [1](3, 21)
    
    return output, H_pdf

def RFStatParams(var,nVar,nCXYZ):
    mean = np.zeros((nVar,nCXYZ))
    std = np.zeros((nVar,nCXYZ))
    for i in range(nVar):
        for j in range(nCXYZ):
            mean[i,j] = np.mean(var[:,j,i])
            std[i,j] = np.std(var[:,j,i])
    
    return mean, std


#==============================================================================
def RFWriteOneRandField(fname2,output,nVar):
    # writing vector of random field to *.csv files
    d = np.savetxt(fname2, output, delimiter = '; ')

def RFWriteOneGPAndRandField(fname2,output,x,nVar):
    # writing vector of random field to *.csv files
    out=[]
    for k in range(len(output)):
      outtmp=[]
      for l in range(len(x[k])):
        outtmp.append(x[k][l])
      for l in range(nVar):
        outtmp.append(output[k][l])
        
      out.append(outtmp)
    d = np.savetxt(fname2, out, delimiter = '; ')


def RFReadOneRandField(fname,nVar,nG):
    randf = np.zeros((nVar,nG)) 
    data = pd.read_csv(fname, delimiter='; ', header=None, engine = 'python') 
    randf = data.to_numpy()         
    print("Read file to generate mesh ", fname)
    return randf

#data = np.random.random((100,5))
#np.savetxt('test.csv', data, fmt = '%.2f', delimiter=',', header = 'c1, c2, c3, c4, c5')
#my_csv = np.loadtxt('./test.csv', delimiter=',')
#my_csv[:5, :]

#==============================================================================
def RFPlotGaussPoints(xG,yG,zG):
    # plotting Gauss points
    fig = plt.figure(figsize = (10.,10.))
    ax = plt.axes(projection='3d')
    ax.grid()
    #plt.rcParams["figure.figsize"] = [8., 6.]
    
    ax.scatter(xG, yG, zG, c = 'r', s = 10)
    ax.set_title('3D Scatter Gauss Points')
    
    # Set axes label
    ax.set_xlabel('x', labelpad=20)
    ax.set_ylabel('y', labelpad=20)
    ax.set_zlabel('z', labelpad=20)
    
    plt.show()
    
def readGP(mechDir):
  foundFile=False
  for file in os.scandir(mechDir):
    if fnmatch.fnmatch(file,'*coordinate_*.csv'):
      print("Read Gauss points in file: ",mechDir+'/'+file.name)
      main_df = pd.DataFrame(pd.read_csv(mechDir+'/'+file.name, delimiter=';', header=None)).to_numpy()
      foundFile=True
      break
  if foundFile==False:
    print("the Gauss points coordinates were not found: need to generate them first")
    exit()
  #print(main_df)
  nGP = len(main_df)                   # no. total Gauss points
  gPCoord = main_df[0:nGP,[1, 2, 3]]    # x,y,z coordinates = (nGTotal,3)
  gPTetNb = main_df[0:nGP,0]    # tetraheadral number of each GP
  gPNb = main_df[0:nGP,4]    # GP number in each Tet

  # determine min, max coordinates
  gCoordMin = np.min(gPCoord,0)
  gCoordMax = np.max(gPCoord,0)
  #print(gCoordMin)
  #print(gCoordMax)
  return gPCoord,gPTetNb,gPNb,gCoordMin,gCoordMax

def writeMeshFile(mechDir,outname,nstep0,gPTetNb,Keys,Fields):
  # Read element information and its average stress from *.msh file
  foundFile=False
  for file in os.scandir(mechDir):
    #print(file)
    if fnmatch.fnmatch(file,'*stress*.msh'):
      inname=mechDir+'/'+file.name
      print("Read Gauss points in file: ",inname)      
      foundFile=True
      break
  if foundFile==False:
    print("the stress file was not found: need to generate them first")
    exit()
  f = open(inname)
  f.readline() # '$MeshFormat\n'
  f.readline() # '2.2 0 8\n'
  f.readline() # '$EndMeshFormat\n'
  f.readline() # '$ElementData'
  f.readline() # '1'
  f.readline() # '"RF"'
  f.readline() # '1'
  f.readline() # '1'
  f.readline() # '4'
  f.readline() # '25'
  f.readline() # '1'
  nel = int(f.readline()) # '768\n'
  f.readline() # '0'
  elem = np.fromfile(f,count=nel*2, sep=" ").reshape((nel,2)) # [iel  sVM]
  f.readline() # 'EndElementData'
  elID = np.reshape((elem[:,0].astype(int)).astype(str),(-1,1))   # (nel,1)
  #elID = np.reshape(elem[:,0].astype(int),(-1,1))     # (nel,1)
  print('Number of Elements: ',nel)
  HeadLine="""$MeshFormat
  2.2 0 8
  $EndMeshFormat"""
  # write *.msh files
  f = open(outname, "w")
  contents = []
  contents.insert(0, HeadLine)
  contents.append("\n")
  contents = "".join(contents)
  f.write(contents)
  for nv in range(len(Keys)):
    print("field ", Keys[nv])
    BeginLine="""$ElementData
1\n\""""+Keys[nv]+"""\"\n1\n"""+str(nstep0)+"""\n4\n"""+str(nstep0)+"""\n1\n"""+str(nel)+"""
0\n""" 
    EndLine = """$EndElementData\n"""
    curEleNb=gPTetNb[0]
    aveval=np.zeros((nel,1))
    npge=0
    avetmp=0.
    index=0
    for k in range(len(gPTetNb)+1):
      if k<len(gPTetNb):
        if gPTetNb[k]==curEleNb:
          npge=npge+1
          avetmp=avetmp+Fields[k][nv]
        else:
          aveval[index]=avetmp/npge
          index=index+1
          avetmp=Fields[k][nv]
          npge=1
          curEleNb=gPTetNb[k]
      else:
        aveval[index]=avetmp/npge
              
    aveval = np.hstack((elID, aveval))              # concatenate elem. ID and the rand. field
    # print(elID,aveval)
    contentB=[]
    contentB.append(BeginLine)
    contentB = "".join(contentB)
    f.write(contentB)
    df_aveval = pd.DataFrame(aveval, index=None, columns=None)  # convert array to data frame
    df_aveval=np.array(df_aveval)
    for v in range(len(df_aveval)): # .to_csv(prefix[i], header=None, index=None, sep=' ')
      #print(df_aveval[v])
      f.write(str(df_aveval[v][0])+' '+str(df_aveval[v][1])+'\n')
    contentE=[]
    contentE.append(EndLine)
    contentE = "".join(contentE)
    f.write(contentE)
    
    #contents = "".join(contents)
    
  f.close()

